
-- original author: unknown
-- modifications by: maighurl@terenas (andr0id)

--[[

General Module Logic

The priority queue of spell cooldowns is updated and then scanned

s1 = the first spell found off cooldown
s2 = the highest priority spell with the lowest cooldown not equal to s1

NOTE: s1 and s2 are guaranteed to be found as inc and inc2 always have 
zero cooldown since they are filler

If it is impossible to cast a spell, such as no shards, the cooldown returned 
is set to 120s cd to push the spell off the list

If we are casting or have recently cast a spell, we put a large CD in the queue
to push this spell out. 

Don't worry, this number will soon be refreshed by 
	UnitBuff(), UnitDebuff() or GetSpellCooldown()

A dummy Incinerate (inc2) is placed at the end of the priority queue to allow
Incinerate to be s1 and s2.

Soulburn Soul Fire casts are delayed until ISF buff is close to dropping in order
to save shards

Empowered Imp Soul Fire casts are queued immediately with a zero cooldown regardless
of time remaining

TODO: hard cast Soul Fire needs to be pushed up based on cast time

common abbreviation

btr buff time remaining
t2t time to target

--]]

-- don't load if class is wrong
local _, class = UnitClass("player")
if class ~= "WARLOCK" then return end

-- load LibRangeCheck, we need this to find distance to target
rc = LibStub:GetLibrary("LibRangeCheck-2.0", true)
if rc then
	--print("LibRangeCheck-2.0 loaded")
end

local GetTime = GetTime
local version = 1.2

local MAX_PRIORITY = 12

-- /clcinfo destro_prio sbsf isf imm con bd cor sf eisf bd2 cb sb inc inc2

local defaults = {
	-- temp db stuff
	priority = {"sbsf", "isf", "imm", "con", "bd", "cor", "sf", "eisf", "bd2", "cb", "sb", "inc", "inc2"},
	clipImmolate = 2.0,
	clipCorruption = 2.0,
	clipISF = 2.0,
	debuffDelay = 0.5,
	version = version,
}

local modName = "_destruction_"
local mod = clcInfo:RegisterClassModule(modName)
local emod = clcInfo.env
local db

local debug = false

-- static spell data
-- i_: spellId, n_: spellName, d_: debuffName, b_: buffName
--------------------------------------------------------------------------------
local i_immolate			= 348		-- Immolate
local i_conflagrate			= 17962		-- Conflagrate
local i_baneofdoom			= 603		-- Bane of Doom
local i_baneofagony			= 980		-- Bane of Agony
local i_shadowflame			= 47897		-- Shadowflame
local i_corruption			= 172		-- Corruption
local i_chaosbolt			= 50796		-- Chaos Bolt
local i_soulburn			= 74434		-- Soulburn
local i_soulfire			= 6353		-- Soul Fire
local i_incinerate			= 29722		-- Incinerate
local i_empoweredimp		= 47220		-- Empowered Imp Talent
local i_shadowburn			= 17877		-- Shadowburn
local i_demonsoul				= 77801   -- Demon Soul 	
local i_timewarp_b			= 80353		-- Time Warp buff
local i_heroism_b				= 32182		-- Heroism buff
local i_bloodlust_b			= 2825		-- Bloodlust buff
local i_felarmor				= 28176		-- Fel Armor
local i_gcd 				= i_immolate

local n_immolate			= GetSpellInfo(i_immolate)
local n_conflagrate			= GetSpellInfo(i_conflagrate)
local n_baneofdoom			= GetSpellInfo(i_baneofdoom)
local n_baneofagony			= GetSpellInfo(i_baneofagony)
local n_shadowflame			= GetSpellInfo(i_shadowflame)
local n_corruption			= GetSpellInfo(i_corruption)
local n_chaosbolt			= GetSpellInfo(i_chaosbolt)
local n_soulburn			= GetSpellInfo(i_soulburn)
local n_soulfire			= GetSpellInfo(i_soulfire)
local n_incinerate			= GetSpellInfo(i_incinerate)
local n_shadowburn			= GetSpellInfo(i_shadowburn)
local n_demonsoul				= GetSpellInfo(i_demonsoul)
local n_felarmor				= GetSpellInfo(i_felarmor)

local b_isf 				= GetSpellInfo(18119)			-- Improved Soul Fire
local b_empoweredimp 		= GetSpellInfo(i_empoweredimp)	-- Empowered Imp
local b_timewarp			= GetSpellInfo(i_timewarp_b)
local b_heroism				= GetSpellInfo(i_heroism_b)
local b_bloodlust			= GetSpellInfo(i_bloodlust_b)

local actionsId = {
	isf						= i_soulfire,
	imm				 		= i_immolate,
	con						= i_conflagrate,
	bd				 		= i_baneofdoom,
	bd2						= i_baneofdoom,
	sf						= i_shadowflame,
	cor						= i_corruption,
	cb						= i_chaosbolt,
	sbsf					= i_soulburn,
	eisf					= i_soulfire,
	inc						= i_incinerate,
	sb						= i_shadowburn,
}

local actionsName = {
	isf						= b_isf, 					-- improved soul fire // if target is above 80%, make sure this buff is up
	imm				 		= n_immolate,
	con						= n_conflagrate,
	bd				 		= n_baneofdoom,
	bd2						= n_baneofdoom .. " (refresh)",
	sf						= n_shadowflame,
	cor						= n_corruption,
	cb						= n_chaosbolt,
	sbsf					= n_soulburn .. " > " .. n_soulfire,
	eisf					= b_empoweredimp .. " > " .. n_soulfire,
	inc						= n_incinerate,
	sb						= n_shadowburn,
}

mod.actionsName = actionsName -- expose for options

-- working priority queue, soulfire priority, skill 1, skill 2
local pq, sfpri, s1, s2

local ef

-- status vars
local _gcd, _ctime, _casting
local _prevUse = {}

-- work vars
local start, duration, expiration

-- initial values to keep calculations from exploding. Computed dynamically hereafter....
local spellspeed = 18				-- warlock spell speed appears to be about 15 - 20 yards/sec
local prevMax = 0
local maxRange = 40
local minRange = 8
local traveltime = (maxRange / spellspeed)
local numShards = 0
local soulburn_cd = 45000

-- buff/debuff time remaining
local cor_btr = 18000
local imm_btr = 15000
local isf_btr = 20000
local bod_btr = 60000
local boa_btr = 24000

-- track target health
local targetHealthPercent = 100

-- track haste to exclude Chaos Bolt when GCD is 1 second
local hasteRating = 0
local prevHasteRating = 0
local skipChaosBolt = false

local actions = {
	---------------------------------------------
	-- Improved Soul Fire	(hard cast Soul Fire)
	---------------------------------------------
	isf = function()
		-- buff lasts 20 seconds, if we are casting or recently cast
		-- just return 20000

		if _casting == n_soulfire then
			return 20000
		end

		-- @hack wait for some time after casting it before deciding buff is not up
		if _prevUse[n_soulfire] then
			if _ctime < _prevUse[n_soulfire] + traveltime + db.debuffDelay then
				return 20000
			else
				_prevUse[n_soulfire] = nil
			end
		end

		-- if the trash target is dying, don't cast, we won't have time to finish the cast
		if UnitLevel("target") >= 0 and targetHealthPercent < 20 then
			return 3000
		end

		-- how long will it take to hard cast this pig?
		local _, _, _, _, _, _, casttime = GetSpellInfo(i_soulfire)
		casttime = casttime / 1000

		-- time to target = casttime + traveltime
		local isf_t2t = casttime + traveltime

		-- adjust so this won't come up before a Soulburn -> Soul Fire
		if numShards > 0 and soulburn_cd == 0 then
			if sfpri["isf"] > sfpri["sbsf"] then
				isf_t2t = traveltime - 0.5
			end
		end
		
		return max(0, isf_btr - (db.clipISF + isf_t2t + _gcd))
	end,

	------------------------------
	-- Soulburn Soul Fire
	------------------------------
	sbsf = function()
		if _casting == n_soulfire then
			return 20000
		end

		-- wait for some time after casting it before deciding buff is not up
		if _prevUse[n_soulfire] then
			if _ctime < _prevUse[n_soulfire] + traveltime + db.debuffDelay then
				return 20000
			else
				_prevUse[n_soulfire] = nil
			end
		end
			
		-- see if we have any shards
		if numShards > 0 then

			local	sbsf_t2t = traveltime

			-- t = time until we must queue sb-sf to refresh ISF on time 			
			local tq = max(0, isf_btr - (db.clipISF + sbsf_t2t + _gcd))
			
			-- if soulburn is off cd before then, we can make it
			if tq >= soulburn_cd then
				return tq
			else
				return 20000
			end
		else
			return 20000			-- no shards, push this off the queue
		end

	end,

	------------------------------
	-- Empowered Imp Soul Fire
	------------------------------
	-- EJ says cast this ASAP
	eisf = function()
		local _, _, _, _, _, _, expiration = UnitBuff("player", b_empoweredimp, nil, "PLAYER")
		if expiration then
			return 0
		end
		return 80000	-- no buff, push this off the queue
	end,

	------------------------------
	-- Immolate	
	------------------------------
	-- debuff lasts 15 seconds, so if we are casting or cast recently 
	-- just return 15000
	
	imm = function()
		-- test if we're casting it
		if _casting == n_immolate then return 15000 end
	
		-- @hack wait for some time after casting it before deciding buff is not up
		if _prevUse[n_immolate] then
			if _ctime < _prevUse[n_immolate] + traveltime + db.debuffDelay then
				return 15000
			else
				_prevUse[n_immolate] = nil
			end
		end

		local _, _, _, _, _, _, casttime = GetSpellInfo(i_immolate)
		casttime = casttime / 1000
		
		-- time to target = casttime + traveltime
		local imm_t2t = casttime + traveltime
		return max(0, imm_btr - (db.clipImmolate + imm_t2t + _gcd))

	end,

	------------------------------
	-- Conflagrate	
	------------------------------
	con = function()
	
		-- target must have Immolate debuff for this to be valid
		local _, _, _, _, _, _, expiration = UnitDebuff("target", n_immolate, nil, "PLAYER")

		-- assume it is valid if Imolate is casting or recently cast
		if _casting == n_immolate then
			expiration = 1
		end
		
		if _prevUse[n_immolate] then
			if _ctime < _prevUse[n_immolate] + traveltime + db.debuffDelay then
				expiration = 1
			end
		end

		if expiration then
			start, duration = GetSpellCooldown(i_conflagrate)
			local con_cd = (start + duration - _ctime - _gcd)
			if con_cd  > 0 then
				return con_cd
			end
			return 0
		end
		
		-- target doesn't have debuff, push this off the queue
		return 120000
	end,
	
	-------------------------------------
	-- Bane of Doom	
	-------------------------------------
	bd = function()
		-- debuff lasts 60 seconds, if recently cast, wait for debuff to show
		if _prevUse[n_baneofdoom] then
			if _ctime < _prevUse[n_baneofdoom] + db.debuffDelay then
				return 60000
			else
				_prevUse[n_baneofdoom] = nil
			end
		end
	  
		-- if we have BoA up, this is probaby a short fight, so ignore BoD for now
		if boa_btr > 0 then 
			return boa_btr
		end

		-- if Bane of Doom is greater than the clip, push it to bd2 for refresh
		if bod_btr >= 5 then
		  	return 60000
		else
			return 0
		end
	end,

	-------------------------------------
	-- Bane of Doom	
	-- refresh at lower priority
	-------------------------------------
	bd2 = function()
		-- debuff lasts 60 seconds, if recently cast, wait for debuff to show
		if _prevUse[n_baneofdoom] then
			if _ctime < _prevUse[n_baneofdoom] + db.debuffDelay then
				return 60000
			else
				_prevUse[n_baneofdoom] = nil
			end
		end

		-- if we have BoA up, this is probaby a short fight, so ignore BoD for now
		if boa_btr > 0 then 
			return boa_btr
		end

		-- if Bane of Doom is below clip reminder and > 5 sec, refresh at low pri
		if 15 > bod_btr and bod_btr > 5 then
		  	return 0
		else
			return 60000
		end
	end,

	------------------------------
	-- Shadowflame
	------------------------------
	sf = function()
		if minRange <= 8 then
			start, duration = GetSpellCooldown(i_shadowflame)
			local sf_cd = max(0, start + duration - _ctime - _gcd)
			return sf_cd
		end
		return 12000
	end,

	------------------------------
	-- Corruption
	------------------------------
	-- debuff lasts 18 seconds, if recenly cast, return 18000
	cor = function()

		if _prevUse[n_corruption] then
			if _ctime < _prevUse[n_corruption] + db.debuffDelay then
				return 18000
			else
				_prevUse[n_corruption] = nil
			end
		end
	
		if cor_btr <= db.clipCorruption then
			return 0
		else
			return cor_btr
		end

	end,

	------------------------------
	-- Chaos Bolt	
	------------------------------
	cb = function()
		-- if casting, return 12 seconds
		if _casting == n_chaosbolt then return 12000 end

		-- if Heroism, Time Warp or Bloodlust
		if skipChaosBolt == true then 
			return 12000 
		end

		start, duration = GetSpellCooldown(i_chaosbolt)
		local cb_cd = max(0, start + duration - _ctime - _gcd)
		return cb_cd
	end,

	------------------------------
	-- Shadowburn
	------------------------------
	sb = function()
		if UnitExists("target") and not UnitIsDead("target") then
			if 20 > targetHealthPercent and targetHealthPercent > 0 then
				start, duration = GetSpellCooldown(i_shadowburn)
				if start == 0 then
					return 0
				end
			end
		end
		return 60000
	end,

	------------------------------
	-- Incinerate - ready when you are
	------------------------------
	inc = function()
		return 0
	end,
	
	------------------------------
	-- Incinerate (dummy for s2)
	------------------------------
	inc2 = function()
	------------------------------
		return 0
	end,
}


------------------------------------------------------------
-- DestructionRotation
------------------------------------------------------------
local function DestructionRotation()
	
	_ctime = GetTime()
	local endTime

	_casting, _, _, _, _, endTime = UnitCastingInfo("player")
	if _casting then
		-- casting a non instant spell	
		_gcd = endTime / 1000 - _ctime
	else
		-- instant or no cast
		start, duration = GetSpellCooldown(i_gcd)
		if duration > 0 then
			_gcd = start + duration - _ctime
		else
			_gcd = 0
		end
	end

	-- get all this info here once on each iteration for efficiency
	------------------------------------------------------------------------------

	-- if we have a target, get some info about it
	if UnitExists("target") then

		targetHealthPercent = 100 * (UnitHealth("target") / UnitHealthMax("target"))
		--print("Target Health: ", targetHealthPercent)
		
		minRange, maxRange = rc:GetRange('target')
		-- we'll use the max to always make sure our buffs stay up
		if maxRange == nil then
			maxRange = 40
		end
	else	-- plug in some default values
		targetHealthPercent = 100
		maxRange = 40
		minRange = 0
	end

	-- See if we have Heroism, Bloodlust or Time Warp, if so we should skip Chaos Bolt
	local _, _, _, _, _, _, HeroismExp = UnitBuff("player", b_heroism, nil, nil)
	local _, _, _, _, _, _, TimeWarpExp = UnitBuff("player", b_timewarp, nil, nil)
	local _, _, _, _, _, _, BloodLustExp = UnitBuff("player", b_bloodlust, nil, nil)

	if HeroismExp or TimeWarpExp or BloodLustExp then
		skipChaosBolt = true
	else
		skipChaosBolt = false
	end	

	-- update spell travel time
	-- spells seem to accelerate a bit, so the multiplier accounts for that
	traveltime = (maxRange / spellspeed) + ((40/maxRange) * 0.2)
	--print("minRange:",minRange,"maxRange",maxRange,"traveltime:", traveltime)

	-- see if we have any shards
	numShards = UnitPower("player", SPELL_POWER_SOUL_SHARDS)

	-- Soulburn cooldown remaining
	start, duration = GetSpellCooldown(i_soulburn)
	soulburn_cd = max(0, start + duration - _ctime)
	
	-- Improved Soul Fire buff time remaining
	local _, _, _, _, _, _, expiration = UnitBuff("player", b_isf, nil, "PLAYER")
	if expiration then
		isf_btr = expiration - _ctime
	else
		isf_btr = 0
	end

	-- Immolate debuff time remaining
	local _, _, _, _, _, _, expiration = UnitDebuff("target", n_immolate, nil, "PLAYER")
	if expiration then
		imm_btr = expiration - _ctime
	else
		imm_btr = 0
	end
	
	-- Corruption debuff time remaining
	local _, _, _, _, _, _, expiration = UnitDebuff("target", n_corruption, nil, "PLAYER")
	if expiration then
		cor_btr = expiration - _ctime
	else
		cor_btr = 0
	end

	-- Band of Doom debuff time remaining
	local _, _, _, _, _, _, expiration = UnitDebuff("target", n_baneofdoom, nil, "PLAYER")
	if expiration then
		bod_btr = expiration - _ctime
	else
		bod_btr = 0
	end

	-- Band of Agony debuff time remaining
	local _, _, _, _, _, _, expiration = UnitDebuff("target", n_baneofagony, nil, "PLAYER")
	if expiration then
		boa_btr = expiration - _ctime
	else
		boa_btr = 0
	end

	-- get cooldowns
	------------------------------------------------------------------------------
	for i, v in ipairs(pq) do
		v.cd = actions[v.alias](i)
	end
	
	-- s1 = first 0 found, s2 = lowest cd ~= s1 
	------------------------------------------------------------------------------
	local sel = 1
	local lowcd = 1000
	local lowid = i_incinerate
	local firstfound = false
		
	for i = 1, #pq do
		-- find first spell off CD
		if ((pq[i].cd == 0) and (firstfound ~= true)) then
			sel = i
			i = i + 1	-- s2 can't be this one, so increment
			firstfound = true
		end
		if (( pq[i] ~= nil) and (pq[i].cd  < lowcd)) then
			lowcd = pq[i].cd
			lowid = pq[i].id
		end
	end

	-- set up initial sequence if everything is off cooldown so icons display correctly
	if sel == 1 and lowcd == 1000 then
		lowid = pq[sel + 1]
	end

	s1 = pq[sel].id
	s2 = lowid
	
	return true
end

------------------------------------------------------------
-- IconDestruction1
------------------------------------------------------------
function emod.IconDestruction1()
	if notinit then DoInit() end
	if DestructionRotation() then
		return emod.IconSpell(s1, true)
	end
end

------------------------------------------------------------
-- IconDestruction2
------------------------------------------------------------
function emod.IconDestruction2()
	if notinit then DoInit() end
	if DestructionRotation() then
		return emod.IconSpell(s2, true)
	end
end

------------------------------------------------------------
-- IconShadowflame
------------------------------------------------------------
function emod.IconShadowflame()
	if notinit then DoInit() end
	if DestructionRotation() then
		if UnitExists("target") and UnitCanAttack("player", "target") and (minRange <= 8) then
			return emod.IconSpell(i_shadowflame, true)
		else
			return emod.IconSpell(0, nil)
		end
	end
end

------------------------------------------------------------
-- IconShadowburn
------------------------------------------------------------
function emod.IconShadowburn()
	if notinit then DoInit() end
	if DestructionRotation() then
		if UnitExists("target") and not UnitIsDead("target") then
			if 20 > targetHealthPercent and targetHealthPercent > 0 then
				start, duration = GetSpellCooldown(i_shadowburn)
				if start + duration - _ctime > 5 then
					return emod.IconSpell(0, nil)
				else				
					return emod.IconSpell(i_shadowburn, true)
				end
			else
				return emod.IconSpell(0, nil)
			end
		end
	end
end

------------------------------------------------------------
-- IconDemonsoul
------------------------------------------------------------
-- todo mess with alpha here to show dimmed for last few seconds of cooldown
function emod.IconDemonsoul()
	if notinit then DoInit() end
	if DestructionRotation() then
			start, duration = GetSpellCooldown(i_demonsoul)
			if start + duration - _ctime > 5 then			
				return emod.IconSpell(0, nil)
			else
				return emod.IconSpell(i_demonsoul, true)
		end
	end
end

------------------------------------------------------------
-- IconFelarmor
------------------------------------------------------------
function emod.IconFelarmor()
	if notinit then DoInit() end
	if DestructionRotation() then
		local _,_,_,_,_,_, exp = UnitBuff("player", n_felarmor)
		if exp then
			return emod.IconSpell(0, nil)
		else
			return emod.IconSpell(i_felarmor, true)
		end
	end
end

------------------------------------------------------------
-- FrameOnEvent
------------------------------------------------------------
local function FrameOnEvent(self, event, unit, spell)
	if unit == "player" then
		_prevUse[spell] = GetTime()
	end
end

-- Set up event frame
local	ef = CreateFrame("Frame", nil, UIParent)
ef:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
ef:SetScript("OnEvent", FrameOnEvent)


local notinit = true
------------------------------------------------------------
-- ExecCleanup
------------------------------------------------------------
local function ExecCleanup()
	ef:UnregisterAllEvents()
	notinit = true
end

------------------------------------------------------------
-- DoInit
------------------------------------------------------------
local function DoInit()
	emod.___e.ExecCleanup = ExecCleanup
	notinit = false
end

------------------------------------------------------------
-- UpdatePriorityQueue
------------------------------------------------------------
function mod.UpdatePriorityQueue()
	pq = {}
	sfpri = {}
	local check = {} -- used to check for duplicates
	for k, v in ipairs(db.priority) do
		if not check[v] then
			if v ~= "none" and actionsId[v] then
				pq[#pq + 1] = { alias = v, id = actionsId[v] }
				check[v] = true
				if v == "isf" or v == "sbsf" then
					sfpri[v] = #pq
				end
			end
		end
	end
	pq[#pq + 1] = { alias = "inc2", id = i_incinerate }
end

------------------------------------------------------------
-- OnInitialize
------------------------------------------------------------
function mod.OnInitialize()
	db = clcInfo:RegisterClassModuleDB(modName, defaults)
	if db then
		if not db.version or db.version < version then
			clcInfo.AdaptConfigAndClean(modName, db, defaults)
			db.version = version
		end

		if debug == true then		
			print("clcInfo Destruction Warlock module version:", db.version)
			print("Immolate clip:", db.clipImmolate)
			print("Corruption clip:", db.clipCorruption)
			print("ISF clip:", db.clipISF)
			print("debuff delay:", db.debuffDelay)
			
			local pri = "pri: "
			for k,v in ipairs(db.priority) do
				pri = pri .. v .. " "
			end
			print(pri)
		end
			
		mod.UpdatePriorityQueue()
	end
end

-- pass filler order from command line
-- intended to be used in macros
local function CmdDestructionPriority(args)
	-- add args to options
	local num = 0
	for i, arg in ipairs(args) do
		if actionsName[arg] then
			if num < MAX_PRIORITY then
				num = num + 1
				db.priority[num] = arg
			else
				print("too many priority specified, max is " .. MAX_PRIORITY)
			end
		else
			-- inform on wrong arguments
			print(arg .. " not found")
		end
	end
	
	-- none on the rest
	if num < MAX_PRIORITY then
		for i = num + 1, MAX_PRIORITY do
			db.priority[i] = "none"
		end
	end

	-- redo queue
	mod.UpdatePriorityQueue()
	
	-- update the options window
	clcInfo:UpdateOptions()
end
-- register for slashcmd
clcInfo.cmdList["destro_prio"] = CmdDestructionPriority

-- edit options from command line
local function CmdDestructionOptions(args)
	if not args[1] or not args[2] then
		print("format: /clcInfo destro_opt option value")
		return
	end
	
	if args[1] == "clipimmolate" then
		db.clipImmolate = tonumber(args[2]) or defaults.clipImmolate
	elseif args[1] == "clipcorruption" then
		db.clipCorruption = tonumber(args[2]) or defaults.clipCorruption
	elseif args[1] == "clipisf" then
		db.clipISF = tonumber(args[2]) or defaults.clipISF
	elseif args[1] == "debuffdelay" then
		db.debuffDelay = tonumber(args[2]) or defaults.debuffDelay
	else
		print("valid options: debuffdelay, clipISF, clipimmolate, clipcorruption")
	end
	
	clcInfo:UpdateOptions()
end
clcInfo.cmdList["destro_opt"] = CmdDestructionOptions

